import re
from functools import cache, partial
from itertools import cycle
from math import lcm
from operator import methodcaller as mc

@cache
def prep(data):
    directions, map_str = data.split("\n\n")
    map_nodes = {a: b for a, *b in map(partial(re.findall, r"\w+"), map_str.split("\n"))}
    return directions, map_nodes

def travel(data, start, end):
    directions, map_nodes = prep(data)
    directions = cycle(directions)
    cur_node = start
    steps = 0
    while not cur_node.endswith(end):
        steps += 1
        cur_node = map_nodes[cur_node][int(next(directions) == "R")]
    return steps

part_a = partial(travel, start="AAA", end="ZZZ")
part_b = lambda d: lcm(*map(partial(travel, d, end="Z"), filter(mc("endswith", "A"), prep(d)[1])))

if __name__ == "__main__":
    # from aocd import get_data
    from aoc_gliech.util import data
    print([d[-1] for d in data()])
    test = """LR

11A = (11B, XXX)
11B = (XXX, 11Z)
11Z = (11B, XXX)
22A = (22B, XXX)
22B = (22C, 22C)
22C = (22Z, 22Z)
22Z = (22B, 22B)
XXX = (XXX, XXX)"""



import re

def prep(data):
    return data.split("\n")

def part_a(data):
    return sum(int((d:=re.findall(r"\d", line))[0]+d[-1]) for line in prep(data))

def part_b(data):
    match = ["\\d", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    return sum((f:=lambda p: int(n) if len(n:=re.match(f"(?:.*{p})({"|".join(match)})", l).group(1))==1 else match.index(n))("?")*10+f("") for l in prep(data))

if __name__ == "__main__":
    from aoc_gliech import solve
    custom_data = """
    two1nine6
    eightwo77eighthree
    abcone2three7xyz
    xtwone34four
    4nineeightseven2
    zoneight234
    7pqrstsixteen
    """
    solve(part_a, part_b, data=custom_data)



from functools import reduce
from itertools import starmap
from operator import mul
from math import ceil, floor, sqrt

def race(time, distance):
    x = sqrt(time**2 - 4*distance)
    min_win = floor((time-x)/2 + 1)
    max_win = ceil((time+x)/2 - 1)
    return max_win-min_win+1

def part_a(data):
    d = zip(*(map(int, v.split()[1:]) for v in data.split("\n")))
    return reduce(mul, starmap(race, d))

def part_b(data):
    d = [int("".join(v.split()[1:])) for v in data.split("\n")]
    return race(*d)

if __name__ == "__main__":
    from aocd import get_data
    print(part_b(get_data()))



from collections import Counter

def prep(data):
    return zip(*[map(int, pair.split()) for pair in data.split("\n")])

def part_a(data):
    return sum(abs(a-b) for a, b in zip(*map(sorted, prep(data))))

def part_b(data):
    list_a, list_b = prep(data)
    lookup_table = Counter(list_b)
    return sum(lookup_table[num]*num for num in list_a)

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



from functools import reduce
from collections import defaultdict
def prep(data):
    return data.split(",")

def hash_algo(s):
    return reduce(lambda x,y: (x+ord(y))*17%256, s, 0)

def part_a(data):
    return sum(map(hash_algo, prep(data)))

def part_b(data):
    boxes = defaultdict(dict)
    for line in prep(data):
        instruction = line.split("=")
        if len(instruction) == 1:
            label = instruction[0][:-1]
            box = hash_algo(label)
            if label in boxes[box]:
                del boxes[box][label]
        else:
            label = instruction[0]
            box = hash_algo(label)
            boxes[box][label] = int(instruction[1])
    return sum(sum((box+1)*(slot+1)*focal_len for slot, focal_len in enumerate(lenses.values())) for box, lenses in boxes.items())

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



from functools import partial
from itertools import batched, chain, count, starmap
import yaml

def create_cat(cat_str):
    title, *map_strs = cat_str.split('\n')
    src, _, dst = title.split()[0].split('-')
    map_definitions = [map(int, map_str.split()) for map_str in map_strs]
    mps = {range(src_start, src_start+length): dst_start-src_start for dst_start, src_start, length in map_definitions}
    return (src, {'nxt': dst, 'mps': mps})

def create_reverse_cat(cat_str):
    title, *map_strs = cat_str.split('\n')
    src, _, dst = title.split()[0].split('-')
    map_definitions = [map(int, map_str.split()) for map_str in map_strs]
    mps = {range(dst_start, dst_start+length): src_start-dst_start for dst_start, src_start, length in map_definitions}
    return (dst, {'nxt': src, 'mps': mps})

def map_item(maps, item_id):
    for map_range, offset in maps.items():
        if item_id in map_range:
            return item_id+offset
    else:
        return item_id

def part_a(data):
    seeds, *map_strings = data.split("\n\n")
    items = list(map(int, seeds.split()[1:]))
    cats = dict(map(create_cat, map_strings))
    cur_cat = 'seed'
    while cur_cat != 'location':
        cat = cats[cur_cat]
        items = list(map(partial(map_item, cat['mps']), items))
        cur_cat = cat['nxt']
    return min(items)

def get_items(items, cats, start_cat, end_cat):
    for item in items:
        cur_cat = start_cat
        start_item = item
        while cur_cat != end_cat:
            cat = cats[cur_cat]
            item = map_item(cat['mps'], item)
            cur_cat = cat['nxt']
        yield (start_item, item)

def part_b(data):
    seeds, *map_strings = data.split("\n\n")
    seeds = [range(s, s+l) for s, l in batched(map(int, seeds.split()[1:]), 2)]
    cats = dict(map(create_reverse_cat, map_strings))
    test_item = lambda a: any(a[1] in r for r in seeds)
    min_seed = next(filter(test_item ,get_items(count(), cats, 'location', 'seed')))
    return min_seed[0]


if __name__ == "__main__":
    from aocd import get_data
    print(part_b(get_data(year=2023, day=5)))



import numpy as np
from itertools import combinations, starmap
from operator import sub

custom_data = """
        ....
        .#..
        ...#
        ....
        ...#"""

def prep(data):
    return np.array([list(line) for line in data.split('\n')], dtype='U1')

def expand(image):
    for axis in 0, 1:
        empty_spaces = np.where(np.all(image == ".", axis))[0]
        image = np.insert(image, empty_spaces, ".", axis=1-axis)
    return image

def galaxy_distances(image):
    return map(sum, map(np.abs, starmap(sub, combinations(map(np.array, zip(*np.where(image == "#"))), 2))))

def part_a(data):
    return sum(galaxy_distances(expand(prep(data))))

def part_b(data):
    image = prep(data)
    dist_past = galaxy_distances(image)
    dist_now = galaxy_distances(expand(image))
    return sum(starmap(lambda a, b: a+(b-a)*999999, zip(dist_past, dist_now)))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b, data=custom_data)



from functools import cmp_to_key, partial
from collections import defaultdict
from operator import eq, ne


def gen_part(op):
    def part(data):
        rules, jobs = map(partial(str.split, sep="\n"), data.split("\n\n"))

        rule_dict = defaultdict(set)
        for key, val in (map(int, rule.split("|")) for rule in rules):
            rule_dict[val].add(key)

        def rule_test(a, b):
            return 1 if b in rule_dict[a] else -1 if a in rule_dict[b] else 0

        jobs = tuple(list(map(int, job.split(","))) for job in jobs)
        sorted_jobs = map(partial(sorted, key=cmp_to_key(rule_test)), jobs)
        filtered_jobs = (i[0] for i in zip(sorted_jobs, jobs) if op(*i))
        return sum(job[len(job)//2] for job in filtered_jobs)

    return part

part_a = gen_part(eq)
part_b = gen_part(ne)

# original solution for part a without sorting
def part_a_orig(data):
    from itertools import starmap
    from collections import deque

    def prep(data):
        data = zip(map(partial(str.split, sep="\n"), data.split("\n\n")), ("|", ","))
        data = starmap(lambda lns, sep: map(lambda ln: tuple(map(int, ln.split(sep))), lns), data)
        return map(tuple, data)

    def middle_page(job):
        return job[len(job)//2]

    rules, jobs = prep(data)

    rule_dict = defaultdict(set)
    for key, val in rules:
        rule_dict[val].add(key)

    def job_ordered(job):
        job = deque(job)
        for _ in range(len(job)):
            rule = rule_dict[job.popleft()]
            if rule & set(job):
                return False
        return True

    return sum(map(middle_page, filter(job_ordered, jobs)))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b, part_a_orig)



from aocd.get import current_day, most_recent_year
from docopt import docopt
from importlib.metadata import version
from pathlib import Path
from textwrap import dedent

aoc_file = """\
def prep(data):
    return data.split("\\n")

def part_a(data):
    return prep(data)

def part_b(data):
    return None

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)
"""

def create_day_file():
    """
    Create a new file to solve advent of code.

    Usage:
        aoc-scaffold [--day=<number>] [--year=<number>]
        aoc-scaffold (-h | --help)
        aoc-scaffold --version

    Options:
        -d, --day=<number>   Day for the file path and templating values.
        -y, --year=<number>  Year for the file path and templating values.
        --version            Show version info.
        -h --help            Show this help screen.
    """

    arguments = docopt(dedent(create_day_file.__doc__), version=version("advent-of-code-gliech"))
    year = int(arguments.year or most_recent_year())
    day = int(arguments.day or current_day())
    file_path = Path(__file__).parent / f"y{year}" / f"d{day:02d}.py"
    if file_path.exists():
        raise FileExistsError(f"File {file_path} already exists")
    file_path.write_text(aoc_file)
    print(f"Created file at {file_path}")



from itertools import pairwise
from math import copysign
from operator import sub

def prep(data):
    return (list(map(int, line.split())) for line in data.split("\n"))

def test_pair(pair):
    diff = sub(*pair)
    return copysign(1 <= abs(diff) <= 3, diff)

def test_report_a(report):
    return len(report)-1 == abs(sum(map(test_pair, pairwise(report))))

def part_a(data):
    return sum(map(test_report_a, prep(data)))

def test_report_a_iter(report):
    last_num = None
    first_valid_test = None
    for num in report:
        if last_num is not None:
            test = test_pair((last_num, num))
            if test == 0:
                return False
            else:
                if first_valid_test is None:
                    first_valid_test = test
                elif first_valid_test != test:
                    return False
        last_num = num
    return True

def part_a_iter(data):
    sum_safe = 0
    for report in prep(data):
        report = list(report)
        if test_report_a_iter(report):
            sum_safe +=1
    return sum_safe

def test_report_b(report):
    dampener_used = False
    last_num = None
    first_valid_test = None
    for num in report:
        if last_num is not None:
            test = test_pair((last_num, num))
            if test == 0:
                if dampener_used:
                    return False
                else:
                    dampener_used = True
                    continue
            else:
                if first_valid_test is None:
                    first_valid_test = test
                elif first_valid_test != test:
                    if dampener_used:
                        return False
                    else:
                        dampener_used = True
                        continue
        last_num = num
    return True

def part_b(data):
    sum_safe = 0
    for report in prep(data):
        report = list(report)
        if test_report_b(report) or test_report_b(reversed(report)):
            sum_safe +=1

    return sum_safe

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



import importlib
import json
import subprocess
from pathlib import Path

def python(year, day, data):
    puzzle_module_name = f"aoc_gliech.y{year}.d{day:02d}"
    try:
        puzzle_solutions = importlib.import_module(puzzle_module_name)
    except ModuleNotFoundError as e:
        if e.name == puzzle_module_name:
            return None, None
        else:
            raise e
    return puzzle_solutions.part_a(data), puzzle_solutions.part_b(data)

def tofu(year, day, data):
    module_path = Path(__file__).parent / "tofu" / str(year) / f"{day:02d}"
    if not module_path.is_dir():
        return None, None
    if not ( module_path / ".terraform" ).is_dir():
        subprocess.check_output(("tofu", "init"), cwd=module_path)
    tofu_proc_out = subprocess.check_output(
        ("tofu", "apply", "-auto-approve", "-json", "-var", f"puzzle_input='{data}'"),
        cwd=module_path
    )
    # stdout of tofu in json mode should be one object per line
    tofu_messages = (json.loads(line) for line in tofu_proc_out.splitlines())
    # There should be 2 messages of type outputs. One for the plan and one for
    # the apply.
    plan_out, apply_out = filter(lambda a: a["type"]=="outputs", tofu_messages)
    # Remove fluff around the output key-value pairs
    module_outputs = {name: attrs["value"] for name, attrs in apply_out["outputs"].items()}
    return module_outputs.get("solution_a"), module_outputs.get("solution_b")



from functools import cache
from numpy import array as arr, where

def prep(data):
    data = data.split('\n')
    return arr([list(line) for line in data], dtype='U1')

pipe_openings = {
    (0, 1): "7-J",
    (0, -1): "L-F",
    (1, 0): "J|L",
    (-1, 0): "7|F"}

def turn_360_degrees(direction):
    return tuple(x*-1 for x in direction)

@cache
def pipe_in_out(char, in_direction):
    mirror_out_direction = next(k for k,v in pipe_openings.items() if k!=in_direction and char in v)
    return turn_360_degrees(mirror_out_direction)

def get_loop(tiles):
    start_pos = arr(next(zip(*where(tiles == "S"))))
    loop = {tuple(start_pos)}

    for start_direction, allowed_connections in pipe_openings.items():
        current_pos = start_pos+arr(start_direction)
        current_char = tiles[*current_pos]
        if current_char in allowed_connections:
            direction = start_direction
            break

    while current_char != "S":
        loop.add(tuple(current_pos))
        direction = pipe_in_out(current_char, direction)
        current_pos += direction
        current_char = tiles[*current_pos]

    possible_start_pipes = set(pipe_openings[turn_360_degrees(start_direction)])
    possible_end_pipes = set(pipe_openings[direction])
    tiles[*start_pos] = (possible_start_pipes & possible_end_pipes).pop()

    return loop

def part_a(data):
    return len(get_loop(prep(data)))//2

def part_b(data):
    tiles = prep(data)
    loop = get_loop(tiles)
    along_loop = {
            "L": "7",
            "F": "J"
            }
    enclosed = 0

    for i in range(tiles.shape[0]):
        inside_loop = False
        for j, char in enumerate(tiles[i]):
            if (i, j) in loop:
                if char in along_loop.keys():
                    on_loop = char
                elif char in along_loop.values():
                    inside_loop = inside_loop != (along_loop[on_loop] == char)
                elif char == "|":
                    inside_loop = not inside_loop
            elif inside_loop:
                enclosed +=1
    return enclosed

if __name__ == "__main__":
    from aoc_gliech import solve
    test = """
        7-F-7-
        .FJ.|7
        SJL.L7
        |F---J
        LJ..LJ
        """
    solve(part_a, part_b, data=test)



from functools import cache
from math import floor
from operator import and_

def prep(data):
    return [len(and_(*map(set, (side.split() for side in card.split(":")[1].split("|"))))) for card in data.split("\n")]

def part_a(data):
    return sum(floor(2**(card-1)) for card in prep(data))

def part_b_short(data):
    d = prep(data)
    return (f:=lambda i: sum(map(f, range(len(d)) if i<0 else range(i+1, i+1+d[i])))+1)(-1)-1

def part_b_fast(data):
    d = prep(data)

    @cache
    def card(i):
        return sum(map(card, range(len(d)) if i<0 else range(i+1, i+1+d[i])))+1

    return card(-1)-1

part_b = part_b_fast

# def sanity_check(data):
#     return all(len(card)==len(set(card)) for card in chain(*prep(data)))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(prep, sanity_check, part_a, part_b_short, part_b_fast, day=4)



import re
from operator import mul

def part_a(data):
    return sum(mul(*map(int, pair)) for pair in re.findall(r"mul\((\d{1,3}),(\d{1,3})\)", data))

def part_b(data):
    return sum(map(part_a, re.findall(r"(?:do\(\)|^).*?(?:don't\(\)|$)", data, flags=re.DOTALL)))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



from collections import Counter
from operator import call
from functools import partial
from itertools import starmap

# def winnings(counter, card_order, data):
#     tie_value = lambda hand: sum(list(card_order).index(value)<<idx*4 for idx, value in enumerate(reversed(hand)))
#     hand_value = lambda hand: (sum(value**2 for value in counter(hand).values())<<20)+tie_value(hand)
#     hands = (starmap(call, zip((hand_value, int), line.split())) for line in data.split("\n"))
#     return sum((rank+1)*next(bid) for rank, bid in enumerate(sorted(hands, key=next)))

w = lambda c,o,d: sum((r+1)*next(b) for r, b in enumerate(sorted((starmap(call,
                  zip((lambda h: (sum(g**2 for g in c(h).values())<<20)+sum(
                  list(o).index(v)<<i*4 for i, v in enumerate(reversed(h))),
                  int), l.split())) for l in d.split("\n")), key=next)))

def jCounter(hand):
    hand = Counter(hand)
    jokers = hand.pop("J", 0)
    hand[next(iter(hand.most_common()), "X")[0]] += jokers
    return hand

part_a = partial(w, Counter, "23456789TJQKA")
part_b = partial(w, jCounter, "J23456789TQKA")

if __name__ == "__main__":
    from aocd import get_data
    print(part_b(get_data(year=2023, day=7)))



import numpy as np
import re
from functools import partial
from itertools import chain
from scipy.ndimage import generic_filter

def all_directions(solver):
    def decorated(data):
        word_field = np.array([list(line) for line in data.split("\n")], dtype='U1')
        return sum(map(solver, map(partial(np.rot90, word_field), range(4))))
    return decorated

@all_directions
def part_a(word_field):
    diag_max = max(word_field.shape)-4
    diag_range = range(-diag_max, diag_max+1)
    all_lines = chain(word_field, map(word_field.diagonal, diag_range))
    return sum(map(lambda a: len(re.findall("XMAS", ''.join(a))), all_lines))

masmas_shape = [[1, 0, 1],
                [0, 1, 0],
                [1, 0, 1]]
masmas_values = np.frombuffer(b"MSAMS", dtype=np.uint8)

@all_directions
def part_b(word_field):
    word_field.dtype = np.int32
    findings = generic_filter(word_field, partial(np.array_equal, masmas_values),
                              footprint=masmas_shape, mode="constant")
    return int(np.sum(findings))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



from collections import defaultdict
from functools import reduce
from itertools import batched
from operator import mul
import re

def prep(data):
    def group_cubes(cubes):
        groups = defaultdict(int)
        for num, color in cubes: groups[color] = max(int(num), groups[color])
        return groups
    return (group_cubes(batched(re.findall(r"[\d\w]+", game.split(":")[1]), 2)) for game in data.split("\n"))

def part_a(data):
    # input = [[{(c := cubes.split(" "))[1]: int(c[0]) for cubes in draw.split(", ")} for draw in game.split(": ")[1].split("; ")] for game in data.split("\n")]
    # return sum(index+1 for index, game in enumerate(input) if not any(starmap(lambda color, num: num > ["red", "green", "blue"].index(color)+12, chain.from_iterable(map(methodcaller("items"), game)))))
    return sum(index+1 for index, game in enumerate(prep(data)) if all(num < ("red", "green", "blue").index(color)+13 for color, num in game.items()))

def part_b(data):
    return sum(reduce(mul, game.values()) for game in prep(data))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(prep, part_a, part_b)



import numpy as np
import re
from collections import defaultdict
from itertools import starmap
from operator import itemgetter, mul

is_symbol = np.vectorize(lambda a: a!='.' and not a.isdigit())

def potential_parts(data):
    data = data.split('\n')
    arr = np.array([list(line) for line in data], dtype='U1')
    for idx, line in enumerate(data):
        for match in re.finditer(r'\d+', line):
            h_min = max(0, idx-1)
            h_max = min(arr.shape[0], idx+2)
            w_min = max(0, match.start()-1)
            w_max = min(arr.shape[1], match.end()+1)
            neighbourhood = arr[h_min:h_max,w_min:w_max]
            yield int(match.group()), neighbourhood

def part_a(data):
    return sum(part[0] for part in potential_parts(data) if is_symbol(part[1]).any())


def get_potential_gears(data):
    data = data.split('\n')
    arr = np.array([list(line) for line in data], dtype='U1')
    for idx, line in enumerate(data):
        for match in re.finditer(r'\d+', line):
            h_min = max(0, idx-1)
            h_max = min(arr.shape[0], idx+2)
            w_min = max(0, match.start()-1)
            w_max = min(arr.shape[1], match.end()+1)
            neighbourhood = arr[h_min:h_max,w_min:w_max]
            for h, w in zip(*np.where(neighbourhood=='*')):
                yield (h_min+h, w_min+w), int(match.group())

def part_b(data):
    potential_gears = defaultdict(list)
    for loc, part in get_potential_gears(data):
        potential_gears[loc].append(part)
    gears = filter(lambda parts: len(parts)==2, potential_gears.values())
    return sum(starmap(mul, gears))


if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



def prep(data):
    lab = {}
    for idx_x, line in enumerate(data.split("\n")):
        for idx_y, char in enumerate(line):
            coord = complex(idx_x, idx_y)
            lab[coord] = char
            if char == "^":
                vec = coord, -1
    return lab, vec

def run_guard(lab, vec, path):
    while True:
        next_pos = sum(vec)
        if vec in path:
            return True
        path.add(vec)
        if next_pos not in lab:
            return False
        vec = (vec[0], vec[1]*-1j) if lab[next_pos] == "#" else (next_pos, vec[1])

def path_coords(lab, vec):
    path = set()
    run_guard(lab, vec, path=path)
    return set(vec[0] for vec in path)

def block_path(lab, coord):
    lab[coord] = "#"
    return lab

def part_a(data):
    return len(path_coords(*prep(data)))

def part_b_shorter(data):
    lab, vec = prep(data)
    coords = path_coords(lab, vec)
    coords.remove(vec[0])
    return sum(run_guard(block_path(lab.copy(), i), vec, set()) for i in coords)

def part_b_faster(data):
    lab, vec = prep(data)
    path, path_coords, loops = set(), set(), 0
    while True:
        next_pos = sum(vec)
        if next_pos not in lab:
            return loops
        if lab[next_pos] == "." and next_pos not in path_coords:
            lab[next_pos] = "#"
            loops += run_guard(lab, vec, path.copy())
            lab[next_pos] = "."
        path.add(vec)
        path_coords.add(vec[0])
        vec = (vec[0], vec[1]*-1j) if lab[next_pos] == "#" else (next_pos, vec[1])

part_b = part_b_faster

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



from itertools import pairwise, starmap
from operator import add, sub
from functools import partial

# Longsolve
# def prep(d):
#     return [list(map(int, l.split())) for l in d.split("\n")]
#
# def part_a(data):
#     predict = lambda ds: ds[-1]+predict([b-a for a,b in pairwise(ds)]) if any(ds) else 0
#     return sum(map(predict, prep(data)))
#
# def part_b(data):
#     predict = lambda ds: ds[0]-predict([b-a for a,b in pairwise(ds)]) if any(ds) else 0
#     return sum(map(predict, prep(data)))

# Higher level solve written out
# def solver(idx, operation):
#     def solve(data):
#         predict = lambda sequence: operation(sequence[idx], predict([b-a for
#                   a,b in pairwise(sequence)])) if any(sequence) else 0
#         return sum(map(predict, prep(data)))
#     return solve

# Golfed
f = lambda i,o: lambda d: sum(map(p:=lambda s: o(s[i], p([b-a for a,b in
    pairwise(s)])) if any(s) else 0, [list(map(int, l.split())) for l in
    d.split("\n")]))
part_a, part_b = starmap(f, ((-1, add), (0, sub)))

if __name__ == "__main__":
    from aoc_gliech import solve
    solve(part_a, part_b)



import inspect
import re
from aocd.get import current_day, get_data, most_recent_year
from collections.abc import Iterator
# from io import StringIO
from itertools import islice
# from prettyprinter import pformat, cpprint
from pprint import pformat
from shutil import get_terminal_size
from textwrap import dedent, indent
from time import perf_counter
from traceback import format_exception

def _get_calling_file_match(match):
    return int(match.group()) if match is not None else None

def solve(*funcs, day=None, year=None, data=None, format_custom_data=True, preview_iterators=8):
    if data is None:
        calling_file = inspect.stack()[1].filename
        year_exp, day_exp = ("{0}(?!.*{0}.*)".format(pattern) for pattern in (r"20\d\d", r"[12]?\d"))
        year_match = re.search(year_exp, calling_file)
        calling_file = re.sub(year_exp, "XXXX", calling_file)
        day_match = re.search(day_exp, calling_file)
        year = year or _get_calling_file_match(year_match) or most_recent_year()
        day = day or _get_calling_file_match(day_match) or current_day()
        data = get_data(year=year, day=day)
    elif format_custom_data:
        data = dedent(data).strip()

    for idx, func in enumerate(funcs):
        func_name = f"Function {idx+1}" if (n:=getattr(func, "__name__", "<")).startswith("<") else n
        time_start = perf_counter()

        try:
            func_output = func(data)
        except Exception as e:
            func_output = e

        time = (perf_counter() - time_start) * 1000
        output_type = type(func_output).__qualname__
        ltitle = f"{func_name} -> {output_type}:  "
        rtitle = f"  [{time:.2f}ms]"
        term_width = get_terminal_size().columns
        width_avail = term_width - len(ltitle+rtitle)

        if preview_iterators and isinstance(func_output, Iterator):
            if preview_iterators is True:
                func_output = tuple(func_output)
            else:
                output_preview = list(islice(func_output, preview_iterators))
                try:
                    next(func_output)
                    output_preview.append("...")
                except StopIteration:
                    pass
                func_output = tuple(output_preview)

        if isinstance(func_output, str) and "\n" in func_output:
            pass
        elif isinstance(func_output, Exception):
            func_output = "".join(format_exception(func_output, limit=-1))
        else:
            # stream = StringIO()
            # cpprint(func_output, stream, compact=True, max_seq_len = 10)
            # func_output = stream.getvalue().strip()
            func_output = pformat(func_output, compact=True, width=term_width-2)

        if "\n" not in func_output and len(func_output) <= width_avail:
            print(ltitle + func_output.ljust(width_avail) + rtitle)
        else:
            print(f"{ltitle}{" "*width_avail}{rtitle}\n{indent(func_output, "  ")}\n")
